unit DWidget;

interface

uses
  Windows, Messages, SysUtils, Classes, GDIPAPI, DWidgetGraphics;

const
  NullPoint: TPoint = (X: 0; Y: 0);

type
  TWidgetWindow = class;
  
  TWidgetApplication = class
  private
    FHandle: HWND;
    FTerminate: Boolean;
    FWidgets: TList;

    function GetWidget(Index: Integer): TWidgetWindow;
    function GetWidgetCount: Integer;
  public
    constructor Create;
    destructor Destroy; override;

    procedure AddWidget(Widget: TWidgetWindow);
    procedure RemoveWidget(Widget: TWidgetWindow);
    procedure RemoveAllWidgets;
    function WidgetFromName(const Name: String): TWidgetWindow;
    function WidgetFromHWND(hWnd: HWND): TWidgetWindow;
    function IndexOfWidget(Widget: TWidgetWindow): Integer;

    procedure Run;
    procedure Terminate;

    property WidgetCount: Integer read GetWidgetCount;
    property Widget[Index: Integer]: TWidgetWindow read GetWidget; default;
  end;

  TWidgetState = set of (wsCreating, wsDestroying, wsWindowCreated, wsVisible,
    wsEnabled, wsShowInTaskbar, wsActive, wsMouseOn, wsStayTopMost,
    wsLButtonDown, wsRButtonDown, wsMButtonDown);
    TMouseButton = (mbLeft, mbRight, mbMiddle);
  
  TWidgetEventDrawRender = procedure(Sender: TObject; const Render: String;
    var DestRect, SrcRect: TRect; var Result: Boolean) of object;
  TWidgetEventTimer = procedure(Sender: TObject; const Name: String) of object;
  
   TMouseEvent = procedure(Sender: TObject; Button: TMouseButton;
    Shift: TShiftState; X, Y: Integer) of object;
  TMouseMoveEvent = procedure(Sender: TObject; Shift: TShiftState;
    X, Y: Integer) of object;
  TKeyEvent = procedure(Sender: TObject; var Key: Word;
    Shift: TShiftState) of object;
  TKeyPressEvent = procedure(Sender: TObject; var Key: Char) of object;

  PRenderItem = ^TRenderItem;
  TRenderItem = packed record
    Name: String;
    Bitmap: TWidgetBitmap;
  end;

  PTimerItem = ^TTimerItem;
  TTimerItem = packed record
    Name: String;
    Index: Integer;
    Interval: DWord;
    Runing: Boolean;
    OnTimer: TWidgetEventTimer;
  end;

  TWidgetWindow = class
  private
    FName: String;
    FHandle: HWND;
    FWidgets: TList;
    FRect: TRect;
    FState: TWidgetState;
    FParent: TWidgetWindow;
    FBlendFunction: TBlendFunction;
    FDIB: TWidgetDIB;
    FMoveOnMouse: Boolean;
    FMoveOnMouseButton: TMouseButton;
    FMouseDownPt: TPoint;
    FResized: Boolean;
    FMoved: Boolean;
    FRenders: TList;
    FTimers: TList;

    FOnCreate: TNotifyEvent;
    FOnDestroy: TNotifyEvent;
    FOnShow: TNotifyEvent;
    FOnHide: TNotifyEvent;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnMouseMove: TMouseMoveEvent;
    FOnMouseDown: TMouseEvent;
    FOnMouseUp: TMouseEvent;
    FOnClick: TNotifyEvent;
    FOnDblClick: TNotifyEvent;
    FOnActivate: TNotifyEvent;
    FOnDeactivate: TNotifyEvent;
    FOnKeyDown: TKeyEvent;
    FOnKeyUp: TKeyEvent;
    FOnDrawRender: TWidgetEventDrawRender;
    FOnBeforeDrawRender: TNotifyEvent;
    FOnAfterDrawRender: TNotifyEvent;
    
    function GetWidget(Index: Integer): TWidgetWindow;
    function GetWidgetCount: Integer;
    function GetLeft: Integer;
    procedure SetLeft(Value: Integer);
    function GetTop: Integer;
    procedure SetTop(Value: Integer);
    function GetWidth: Integer;
    procedure SetWidth(Value: Integer);
    function GetHeight: Integer;
    procedure SetHeight(Value: Integer);
    function GetWidgetScreenLeft: Integer;
    function GetWidgetScreenTop: Integer;
    function GetEnabled: Boolean;
    procedure SetEnabled(Value: Boolean);
    function GetVisible: Boolean;
    procedure SetVisible(Value: Boolean);
    function GetShowInTaskbar: Boolean;
    procedure SetShowInTaskbar(Value: Boolean);
    function GetStayTopMost: Boolean;
    procedure SetStayTopMost(Value: Boolean);
    function GetClientRect: TRect;
  protected
    procedure WndProc(Msg: DWORD; wParam: WPARAM; lParam: LPARAM; var Result: LRESULT); virtual;

    procedure CreateWindow; virtual;
    procedure DestroyWindow; virtual;

    function RenderIndexOf(const Render: String): Integer; overload;
    function RenderIndexOf(Render: TWidgetBitmap): Integer; overload;
    function TimerIndexOf(const Name: String): Integer;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure AddWidget(Widget: TWidgetWindow);
    procedure RemoveWidget(Widget: TWidgetWindow);
    procedure RemoveAllWidgets;
    function IndexOfWidget(Widget: TWidgetWindow): Integer;

    procedure SetWidgetRect(Left, Top, Right, Bottom: Integer); virtual;
    procedure SetWidgetRect2(Left, Top, Width, Height: Integer); virtual;

    procedure AdjustSize; virtual;
    procedure Show; virtual;
    procedure Hide; virtual;
    procedure Render; virtual;
    procedure Update; virtual;
    procedure Init; virtual;

    function SetRender(const FileName: String; Render: String): TWidgetBitmap; overload;
    function SetRender(Stream: TStream; Render: String): TWidgetBitmap; overload;
    function GetRender(Render: String): TWidgetBitmap; overload;
    function GetRender(Render: TWidgetBitmap): String; overload;
    procedure RemoveRender(Render: String); overload;
    procedure RemoveRender(Render: TWidgetBitmap); overload;
    procedure RemoveAllRenders;
    procedure UseRenderAsSize(Render: TWidgetBitmap); overload;
    procedure UseRenderAsSize(Render: String); overload;

    procedure SetTimer(const Name: String; Interval: DWord; OnTimer: TWidgetEventTimer; Start: Boolean = True); overload;
    procedure SetTimer(const Name: String; Interval: DWord); overload;
    procedure SetTimer(const Name: String; OnTimer: TWidgetEventTimer); overload;
    function GetTimerInterval(const Name: String): DWord;
    function GetTimerStarted(const Name: String): Boolean;
    procedure SetTimerState(const Name: String; Start: Boolean);
    procedure RemoveTimer(const Name: String);
    procedure RemoveAllTimers;
    
    property WidgetName: String read FName write FName;
    property WidgetEnabled: Boolean read GetEnabled write SetEnabled;
    property WidgetVisible: Boolean read GetVisible write SetVisible;
    property WidgetShowInTaskbar: Boolean read GetShowInTaskbar write SetShowInTaskbar;
    property WidgetStayTopMost: Boolean read GetStayTopMost write SetStayTopMost;
    property WidgetMoveOnMouse: Boolean read FMoveOnMouse write FMoveOnMouse;
    property WidgetMoveOnMouseButton: TMouseButton read FMoveOnMouseButton write FMoveOnMouseButton;
    property WidgetState: TWidgetState read FState;
    property WidgetParent: TWidgetWindow read FParent;
    property WidgetRect: TRect read FRect write FRect;
    property WidgetClientRect: TRect read GetClientRect;
    property WidgetLeft: Integer read GetLeft write SetLeft;
    property WidgetTop: Integer read GetTop write SetTop;
    property WidgetWidth: Integer read GetWidth write SetWidth;
    property WidgetHeight: Integer read GetHeight write SetHeight;
    property WidgetScreenLeft: Integer read GetWidgetScreenLeft;
    property WidgetScreenTop: Integer read GetWidgetScreenTop;
    property WidgetHandle: HWND read FHandle;
    property WidgetDIB: TWidgetDIB read FDIB;
    property WidgetCount: Integer read GetWidgetCount;
    property Widget[Index: Integer]: TWidgetWindow read GetWidget; default;
    
    property OnCreate: TNotifyEvent read FOnCreate write FOnCreate;
    property OnDestroy: TNotifyEvent read FOnDestroy write FOnDestroy;
    property OnShow: TNotifyEvent read FOnShow write FOnShow;
    property OnHide: TNotifyEvent read FOnHide write FOnHide;
    property OnMouseMove: TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
    property OnActivate: TNotifyEvent read FOnActivate write FOnActivate;
    property OnDeactivate: TNotifyEvent read FOnDeactivate write FOnDeactivate;
    property OnKeyDown: TKeyEvent read FOnKeyDown write FOnKeyDown;
    property OnKeyUp: TKeyEvent read FOnKeyUp write FOnKeyUp;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnDrawRender: TWidgetEventDrawRender read FOnDrawRender write FOnDrawRender;
    property OnBeforeDrawRender: TNotifyEvent read FOnBeforeDrawRender write FOnBeforeDrawRender;
    property OnAfterDrawRender: TNotifyEvent read FOnAfterDrawRender write FOnAfterDrawRender;
  end;

var
  WidgetApplication: TWidgetApplication;

implementation

var
  WidgetHideWindow: HWND;
  WidgetWindowClass: TWndClass = (
    style: CS_HREDRAW or CS_VREDRAW or CS_DBLCLKS;
    cbClsExtra: 0;
    cbWndExtra: 0;
    hInstance: 0;
    hIcon: 0;
    hCursor: 0;
    hbrBackground: 0;
    lpszMenuName: NIL;
    lpszClassName: 'TWidgetWindow');

function WidgetWindowProc(hWnd: HWND; Msg: DWORD; wParam: WPARAM; lParam: LPARAM): LRESULT stdcall;
var
  Widget: TWidgetWindow;
begin
  Result := 0;
  Widget := WidgetApplication.WidgetFromHWND(hWnd);
  if Assigned(Widget) then
    Widget.WndProc(Msg, wParam, lParam, Result);
  if (Result = 0) then
    Result := DefWindowProc(hWnd, Msg, wParam, lParam);
end;

// from Forms.pas
function KeyDataToShiftState(KeyData: Longint): TShiftState;
const
  AltMask = $20000000;
{$IFDEF LINUX}
  CtrlMask = $10000000;
  ShiftMask = $08000000;
{$ENDIF}
begin
  Result := [];
  if GetKeyState(VK_SHIFT) < 0 then Include(Result, ssShift);
  if GetKeyState(VK_CONTROL) < 0 then Include(Result, ssCtrl);
  if KeyData and AltMask <> 0 then Include(Result, ssAlt);
{$IFDEF LINUX}
  if KeyData and CtrlMask <> 0 then Include(Result, ssCtrl);
  if KeyData and ShiftMask <> 0 then Include(Result, ssShift);
{$ENDIF}
end;

function KeyboardStateToShiftState(const KeyboardState: TKeyboardState): TShiftState; overload;
begin
  Result := [];
  if KeyboardState[VK_SHIFT] and $80 <> 0 then Include(Result, ssShift);
  if KeyboardState[VK_CONTROL] and $80 <> 0 then Include(Result, ssCtrl);
  if KeyboardState[VK_MENU] and $80 <> 0 then Include(Result, ssAlt);
  if KeyboardState[VK_LBUTTON] and $80 <> 0 then Include(Result, ssLeft);
  if KeyboardState[VK_RBUTTON] and $80 <> 0 then Include(Result, ssRight);
  if KeyboardState[VK_MBUTTON] and $80 <> 0 then Include(Result, ssMiddle);
end;

function KeyboardStateToShiftState: TShiftState; overload;
var
  KeyState: TKeyBoardState;
begin
  GetKeyboardState(KeyState);
  Result := KeyboardStateToShiftState(KeyState);
end;

{ TWidgetApplication }

constructor TWidgetApplication.Create;
begin
  inherited;
  FWidgets := TList.Create;
  FTerminate := False;
  FHandle := CreateWindow(WidgetWindowClass.lpszClassName, '',
    WS_POPUP, 0, 0, 0, 0, 0, 0, HInstance, NIL);
end;

destructor TWidgetApplication.Destroy;
begin
  DestroyWindow(FHandle);
  RemoveAllWidgets;
  FWidgets.Free;
  inherited;
end;

function TWidgetApplication.GetWidget(Index: Integer): TWidgetWindow;
begin
  Result := FWidgets[Index];
end;

function TWidgetApplication.GetWidgetCount: Integer;
begin
  Result := FWidgets.Count;
end;

procedure TWidgetApplication.AddWidget(Widget: TWidgetWindow);
begin
  FWidgets.Add(Widget);
  if (Length(Widget.WidgetName) = 0) then
    Widget.WidgetName := Format('WidgetWindow%d', [FWidgets.Count]);
  if not Assigned(Widget.WidgetParent) then
    Widget.CreateWindow;
end;

procedure TWidgetApplication.RemoveWidget(Widget: TWidgetWindow);
begin
  FWidgets.Remove(Widget);
  if not (wsDestroying in Widget.WidgetState) then
    Widget.Free;
end;

procedure TWidgetApplication.RemoveAllWidgets;
var
  i: Integer;
  Widget: TWidgetWindow;
begin
  for i:= FWidgets.Count - 1 downto 0 do
  begin
    Widget := FWidgets[i];
    FWidgets.Delete(i);
    if not (wsDestroying in Widget.WidgetState) then
      Widget.Free;
  end;
end;

function TWidgetApplication.IndexOfWidget(Widget: TWidgetWindow): Integer;
begin
  Result := FWidgets.IndexOf(Widget);
end;

function TWidgetApplication.WidgetFromName(const Name: String): TWidgetWindow;
var
  i: Integer;
begin
  Result := NIL;
  for i:= FWidgets.Count - 1 downto 0 do
  if (TWidgetWindow(FWidgets[i]).WidgetName = Name) then
  begin
    Result := FWidgets[i];
    Break;
  end;
end;

function TWidgetApplication.WidgetFromHWND(hWnd: HWND): TWidgetWindow;

function FindInWidget(Widget: TWidgetWindow): TWidgetWindow;
var
  i: Integer;
begin
  Result := NIL;
  for i:= 0 to Widget.WidgetCount - 1 do
  if (wsWindowCreated in Widget.Widget[i].FState) then
  begin;
    if (Widget.Widget[i].FHandle = hwnd) then
      Result := Widget.Widget[i]
    else
      Result := FindInWidget(Widget.Widget[i]);
    if (Result <> NIL) then
      Break;
  end;
end;

var
  i: Integer;
begin
  Result := NIL;
  for i:= 0 to WidgetCount - 1 do
  if (wsWindowCreated in Widget[i].FState) then
  begin;
    if (Widget[i].FHandle = hwnd) then
      Result := Widget[i]
    else
      Result := FindInWidget(Widget[i]);
    if (Result <> NIL) then
      Break;
  end;
end;

procedure TWidgetApplication.Run;

var
  Msg: TMsg;

function ProcessMessage: Boolean;
begin
  Result := PeekMessage(Msg, 0, 0, 0, PM_REMOVE);
  if Result then
    if (Msg.message <> WM_QUIT) then
    begin
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end else
      FTerminate := True;
end;

begin
  while not FTerminate do    
    if not ProcessMessage then WaitMessage;
end;

procedure TWidgetApplication.Terminate;
begin
  PostQuitMessage(0);
end;

{ TWidgetWindow }

constructor TWidgetWindow.Create;
begin
  inherited;
  FWidgets := TList.Create;
  FRect.Left := 0;
  FRect.Top := 0;
  FRect.Right := 200;
  FRect.Bottom := 300;
  FState := [wsEnabled];
  FParent := NIL;
  FRenders := TList.Create;
  FMoveOnMouse := False;
  FMoveOnMouseButton := mbLeft;
  FResized := False;
  FMoved := False;
  FTimers := TList.Create;
  with FBlendFunction do
  begin
    BlendOp := AC_SRC_OVER;
    BlendFlags := 0;
    AlphaFormat := AC_SRC_ALPHA;
    SourceConstantAlpha := 255;
  end;
end;

destructor TWidgetWindow.Destroy;
begin
  Include(FState, wsDestroying);
  DestroyWindow;
  if Assigned(FParent) then
    FParent.RemoveWidget(Self);
  RemoveAllWidgets;
  FWidgets.Free;
  RemoveAllRenders;
  FRenders.Clear;
  RemoveAllTimers;
  FTimers.Free;
  inherited;
end;

function TWidgetWindow.GetWidget(Index: Integer): TWidgetWindow;
begin
  Result := FWidgets[Index];
end;

function TWidgetWindow.GetWidgetCount: Integer;
begin
  Result := FWidgets.Count;
end;

procedure TWidgetWindow.AddWidget(Widget: TWidgetWindow);
begin
  FWidgets.Add(Widget);
  if (Length(Widget.WidgetName) = 0) then
    Widget.WidgetName := Format('WidgetWindow%d', [FWidgets.Count]);
  Widget.FParent := Self;
  Widget.CreateWindow;
end;

procedure TWidgetWindow.RemoveWidget(Widget: TWidgetWindow);
begin
  FWidgets.Remove(Widget);
  if not (wsDestroying in Widget.WidgetState) then
    Widget.Free;
end;

procedure TWidgetWindow.RemoveAllWidgets;
var
  i: Integer;
  Widget: TWidgetWindow;
begin
  for i:= FWidgets.Count - 1 downto 0 do
  begin
    Widget := FWidgets[i];
    FWidgets.Delete(i);
    if not (wsDestroying in Widget.WidgetState) then
      Widget.Free;
  end;
end;

function TWidgetWindow.IndexOfWidget(Widget: TWidgetWindow): Integer;
begin
  Result := FWidgets.IndexOf(Widget);
end;

procedure TWidgetWindow.WndProc(Msg: DWORD; wParam: WPARAM; lParam: LPARAM; var Result: LRESULT);
var
  Key: Word;
  tme: TTrackMouseEvent;
  Pt: TPoint;
  i: Integer;
begin
  case Msg of

    WM_TIMER:
    for i:= 0 to FTimers.Count - 1 do
    if (PTimerItem(FTimers[i])^.Index = wParam) then
    begin
      if Assigned(PTimerItem(FTimers[i])^.OnTimer) then
        PTimerItem(FTimers[i])^.OnTimer(Self, PTimerItem(FTimers[i])^.Name);
      Break;
    end;

    WM_DESTROY:
    begin
      FDIB.Free;
      Exclude(FState, wsWindowCreated);
      Exclude(FState, wsActive);
      Exclude(FState, wsMouseOn);
      Exclude(FState, wsLButtonDown);
      Exclude(FState, wsRButtonDown);
      Exclude(FState, wsMButtonDown);
      if Assigned(FOnDestroy) then
        FOnDestroy(Self);
      if not (wsDestroying in FState) then
        Free;
    end;

    WM_ACTIVATE:
    if (WParam = WA_INACTIVE) then
    begin
      Exclude(FState, wsActive);
      if Assigned(FOnDeactivate) then
        FOnDeactivate(Self);
    end else
    begin
      Include(FState, wsActive);
      if Assigned(FOnActivate) then
        FOnActivate(Self);
    end;            

    WM_KEYDOWN:
    begin
      Key := wParam;
      if Assigned(FOnKeyDown) then
        FOnKeyDown(Self, Key, KeyDataToShiftState(LParam));
    end;

    WM_KEYUP:
    begin
      Key := wParam;
      if Assigned(FOnKeyUp) then
        FOnKeyUp(Self, Key, KeyDataToShiftState(LParam));
    end;

    WM_MOUSEMOVE:
    begin
      tme.cbSize := SizeOf(TTrackMouseEvent);
      tme.dwFlags := TME_LEAVE ;
      tme.hwndTrack := FHandle;
      tme.dwHoverTime := HOVER_DEFAULT;
      TrackMouseEvent(tme);

      if not (wsMouseOn in FState) then
      begin
        Include(FState, wsMouseOn);
        if Assigned(FOnMouseEnter) then
          FOnMouseEnter(Self);
      end;
      
      Pt.X := LOWORD(lParam);
      Pt.Y := HIWORD(lParam);
      if Assigned(FOnMouseMove) then
        FOnMouseMove(Self, KeyboardStateToShiftState, Pt.X, Pt.Y);

      if not Assigned(FParent) and FMoveOnMouse and
        (((FMoveOnMouseButton = mbLeft) and (wsLButtonDown in FState)) or
        ((FMoveOnMouseButton = mbMiddle) and (wsMButtonDown in FState)) or
        ((FMoveOnMouseButton = mbRight) and (wsRButtonDown in FState))) then
      begin
        SetWidgetRect2(
          FRect.Left - FMouseDownPt.X + Pt.X,
          FRect.Top - FMouseDownPt.Y + Pt.Y,
          FRect.Right - FRect.Left, FRect.Bottom - FRect.Top);
      end;
    end;

    WM_MOUSELEAVE:
    begin
      Exclude(FState, wsMouseOn);
      Exclude(FState, wsLButtonDown);
      Exclude(FState, wsRButtonDown);
      Exclude(FState, wsMButtonDown);
      if Assigned(FOnMouseLeave) then
        FOnMouseLeave(Self);
    end;

    WM_LBUTTONDOWN:
    begin
      Include(FState, wsLButtonDown);
      Pt.X := LOWORD(lParam);
      Pt.Y := HIWORD(lParam);
      FMouseDownPt := Pt;
      if Assigned(FOnMouseDown) then
        FOnMouseDown(Self, mbLeft, KeyboardStateToShiftState, Pt.X, Pt.Y);
    end;

    WM_MBUTTONDOWN:
    begin
      Include(FState, wsMButtonDown);
      Pt.X := LOWORD(lParam);
      Pt.Y := HIWORD(lParam);
      FMouseDownPt := Pt;
      if Assigned(FOnMouseDown) then
        FOnMouseDown(Self, mbMiddle, KeyboardStateToShiftState, Pt.X, Pt.Y);
    end;
    
    WM_RBUTTONDOWN:
    begin
      Include(FState, wsRButtonDown);
      Pt.X := LOWORD(lParam);
      Pt.Y := HIWORD(lParam);
      FMouseDownPt := Pt;
      if Assigned(FOnMouseDown) then
        FOnMouseDown(Self, mbRight, KeyboardStateToShiftState, Pt.X, Pt.Y);
    end;

    WM_LBUTTONUP:
    begin
      if Assigned(FOnClick) and (wsLButtonDown in FState) and (wsActive in FState) then
        FOnClick(Self);
      Exclude(FState, wsLButtonDown);
      Pt.X := LOWORD(lParam);
      Pt.Y := HIWORD(lParam);
      if Assigned(FOnMouseUp) then
        FOnMouseUp(Self, mbLeft, KeyboardStateToShiftState, Pt.X, Pt.Y);
    end;

    WM_MBUTTONUP:
    begin
      if Assigned(FOnClick) and (wsMButtonDown in FState) and (wsActive in FState) then
        FOnClick(Self);
      Exclude(FState, wsMButtonDown);
      Pt.X := LOWORD(lParam);
      Pt.Y := HIWORD(lParam);
      if Assigned(FOnMouseUp) then
        FOnMouseUp(Self, mbMiddle, KeyboardStateToShiftState, Pt.X, Pt.Y);
    end;
    
    WM_RBUTTONUP:
    begin
      if Assigned(FOnClick) and (wsRButtonDown in FState) and (wsActive in FState) then
        FOnClick(Self);
      Exclude(FState, wsRButtonDown);
      Pt.X := LOWORD(lParam);
      Pt.Y := HIWORD(lParam);
      if Assigned(FOnMouseUp) then
        FOnMouseUp(Self, mbRight, KeyboardStateToShiftState, Pt.X, Pt.Y);
    end;
    
    WM_LBUTTONDBLCLK:
    if Assigned(FOnDblClick) then
      FOnDblClick(Self);

    WM_MBUTTONDBLCLK:
    if Assigned(FOnDblClick) then
      FOnDblClick(Self);

    WM_RBUTTONDBLCLK:
    if Assigned(FOnDblClick) then
      FOnDblClick(Self);

  end;
end;

procedure TWidgetWindow.CreateWindow;
const
  FlagEnabled: array[Boolean] of DWORD = (WS_DISABLED, 0);
  FlagShow: array[Boolean] of DWORD = (0, WS_VISIBLE);

begin
  if not (wsWindowCreated in FState) and not (wsCreating in FState) then
  begin
    Include(FState, wsCreating);

    if Assigned(FParent) then
      FHandle := CreateWindowEx(
        WS_EX_LAYERED, WidgetWindowClass.lpszClassName, '',
        WS_POPUP or FlagShow[wsVisible in FState] or FlagEnabled[wsEnabled in FState],
        FParent.WidgetScreenLeft + FRect.Left, FParent.WidgetScreenTop + FRect.Top,
        FRect.Right - FRect.Left, FRect.Bottom - FRect.Top,
        FParent.FHandle, 0, HInstance, NIL)
    else
    begin
      FHandle := CreateWindowEx(
        WS_EX_LAYERED, WidgetWindowClass.lpszClassName, '',
        WS_POPUP or FlagShow[wsVisible in FState] or FlagEnabled[wsEnabled in FState],
        FRect.Left, FRect.Top,
        FRect.Right - FRect.Left, FRect.Bottom - FRect.Top,
        0, 0, HInstance, NIL);
      if not (wsShowInTaskBar in FState) then
        SetWindowLong(FHandle, GWL_HWNDPARENT, WidgetHideWindow);
    end;                           
    SetWindowLong(FHandle, GWL_WNDPROC, Integer(@WidgetWindowProc));
    
    if (wsStayTopMost in FState) then
      SetWindowPos(FHandle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE);

    FDIB := TWidgetDIB.Create(FRect.Right - FRect.Left, FRect.Bottom - FRect.Top);
    FDIB.Fill;

    if Assigned(FOnCreate) then
      FOnCreate(Self);
    
    Exclude(FState, wsCreating);
    Include(FState, wsWindowCreated);
    
    Render;
    Update;
  end;
end;

procedure TWidgetWindow.DestroyWindow;
begin
  if (wsWindowCreated in FState) then
    Windows.DestroyWindow(FHandle);
end;

procedure TWidgetWindow.Init;
begin
  Include(FState, wsDestroying);
  DestroyWindow;          
  Exclude(FState, wsDestroying);
  CreateWindow;
end;

function TWidgetWindow.GetWidgetScreenLeft: Integer;
var
  Pt: TPoint;
begin
  Pt.X := FRect.Left;
  ClientToScreen(FHandle, Pt);
  Result := Pt.X;
end;

function TWidgetWindow.GetWidgetScreenTop: Integer;
var
  Pt: TPoint;
begin
  Pt.Y := FRect.Top;
  ClientToScreen(FHandle, Pt);
  Result := Pt.Y;
end;

function TWidgetWindow.GetLeft: Integer;
begin
  Result := FRect.Left;
end;

function TWidgetWindow.GetTop: Integer;
begin
  Result := FRect.Top;
end;

function TWidgetWindow.GetWidth: Integer;
begin
  Result := FRect.Right - FRect.Left;
end;

function TWidgetWindow.GetHeight: Integer;
begin
  Result := FRect.Bottom - FRect.Top;
end;

procedure TWidgetWindow.SetLeft(Value: Integer);
begin
  SetWidgetRect(Value, FRect.Top, Value + FRect.Right - FRect.Left, FRect.Bottom);
end;

procedure TWidgetWindow.SetTop(Value: Integer);
begin
  SetWidgetRect(FRect.Left, Value, FRect.Right, Value + FRect.Bottom - FRect.Top);
end;

procedure TWidgetWindow.SetWidth(Value: Integer);
begin
  SetWidgetRect(FRect.Left, FRect.Top, FRect.Left + Value, FRect.Bottom);
end;

procedure TWidgetWindow.SetHeight(Value: Integer);
begin
  SetWidgetRect(FRect.Left, FRect.Top, FRect.Right, FRect.Top + Value);
end;

procedure TWidgetWindow.SetWidgetRect(Left, Top, Right, Bottom: Integer);
begin
  FResized := (FRect.Right - FRect.Left <> Right - Left) or
    (FRect.Bottom - FRect.Top <> Bottom - Top);
  FMoved := (FRect.Left <> Left) or (FRect.Top <> Top);
  if not FMoved and not FResized then
    Exit;
  FRect.Left := Left;
  FRect.Top := Top;
  FRect.Right := Right;
  FRect.Bottom := Bottom;
  AdjustSize;
end;

procedure TWidgetWindow.SetWidgetRect2(Left, Top, Width, Height: Integer);
begin
  SetWidgetRect(Left, Top, Left + Width, Top + Height);
end;

procedure TWidgetWindow.AdjustSize;
var
  i: Integer; 
begin
  if Assigned(FParent) then
    SetWindowPos(FHandle, 0, FParent.WidgetLeft + FRect.Left, FParent.WidgetTop + FRect.Top,
      FRect.Right - FRect.Left, FRect.Bottom - FRect.Top, SWP_NOZORDER)
  else
    SetWindowPos(FHandle, 0, FRect.Left, FRect.Top,
      FRect.Right - FRect.Left, FRect.Bottom - FRect.Top, SWP_NOZORDER);
  if (wsStayTopMost in FState) then
    SetWindowPos(FHandle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE);
  for i:= 0 to FWidgets.Count - 1 do
    TWidgetWindow(FWidgets[i]).AdjustSize;
  if FResized then
    Render;
  Update;
  FResized := False;
end;

procedure TWidgetWindow.Render;
var
  i: Integer;
  Result: Boolean;
  DestRect, SrcRect: TRect;
begin
  if (wsWindowCreated in FState) then
  begin
    if Assigned(FOnBeforeDrawRender) then
      FOnBeforeDrawRender(Self);
    FDIB.BrushColor := 0;
    FDIB.Fill;
    for i:= 0 to FRenders.Count - 1 do
    begin
      DestRect := FDIB.Rect;
      SrcRect := PRenderItem(FRenders[i])^.Bitmap.Rect;
      Result := (i = 0);
      if Assigned(FOnDrawRender) then
        FOnDrawRender(Self, PRenderItem(FRenders[i])^.Name, DestRect, SrcRect, Result);
      if Result then
        FDIB.DrawBitmap(DestRect, PRenderItem(FRenders[i])^.Bitmap.Bitmap, SrcRect);
    end;
    if Assigned(FOnAfterDrawRender) then
      FOnAfterDrawRender(Self);
  end;
end;

procedure TWidgetWindow.Update;
var
  TopLeft: TPoint;
  Size: TSize;
begin
  if (wsWindowCreated in FState) and (wsVisible in FState) then
  begin
    TopLeft.X := FRect.Left;
    TopLeft.Y := FRect.Top;
    if Assigned(FParent) then
    begin
      inc(TopLeft.X, FParent.WidgetLeft);
      inc(TopLeft.Y, FParent.WidgetTop);
    end;
    Size.cx := FRect.Right - FRect.Left;
    Size.cy := FRect.Bottom - FRect.Top; 
    UpdateLayeredWindow(FHandle, 0, @TopLeft, @Size, FDIB.DC, @NullPoint, 0, @FBlendFunction, ULW_ALPHA);
  end;
end;

function TWidgetWindow.GetEnabled: Boolean;
begin
  Result := (wsEnabled in FState);
end;

procedure TWidgetWindow.SetEnabled(Value: Boolean);
begin
  if Value then
    Include(FState, wsEnabled)
  else
    Exclude(FState, wsEnabled);
end;

function TWidgetWindow.GetShowInTaskbar: Boolean;
begin
  Result := (wsShowInTaskbar in FState);
end;

procedure TWidgetWindow.SetShowInTaskbar(Value: Boolean);
begin
  if not Assigned(FParent) and (GetShowInTaskbar <> Value) then
  begin
    if Value then
    begin
      Include(FState, wsShowInTaskbar);
      if (wsVisible in FState) then
      begin
        SetWindowLong(FHandle, GWL_HWNDPARENT, 0);
        SetForegroundWindow(FHandle);
      end;
    end else
    begin
      Exclude(FState, wsShowInTaskbar);
      if (wsVisible in FState) then
        SetWindowLong(FHandle, GWL_HWNDPARENT, WidgetHideWindow);
    end;
  end;
end;

function TWidgetWindow.GetStayTopMost: Boolean;
begin
  Result := (wsStayTopMost in FState);
end;

procedure TWidgetWindow.SetStayTopMost(Value: Boolean);
begin
  if not Assigned(FParent) and (GetShowInTaskbar <> Value) then
  begin
    if Value then
    begin
      Include(FState, wsStayTopMost);
      if (wsVisible in FState) then
        SetWindowPos(FHandle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE);
    end else
    begin
      Exclude(FState, wsStayTopMost);
      if (wsVisible in FState) then
        SetWindowPos(FHandle, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE);
    end;
  end;
end;

function TWidgetWindow.GetVisible: Boolean;
begin
  Result := (wsVisible in FState);
end;

procedure TWidgetWindow.SetVisible(Value: Boolean);
var
  i: Integer;
begin
  if (GetVisible <> Value) then
  begin
    if Value then
    begin
      Include(FState, wsVisible);
      if (wsWindowCreated in FState) then
      begin
        AdjustSize;
        ShowWindow(FHandle, SW_SHOW);
        if Assigned(FOnShow) then
          FOnShow(Self);
      end;
    end else
    begin
      Exclude(FState, wsVisible);
      if (wsWindowCreated in FState) then
      begin
        ShowWindow(FHandle, SW_HIDE);
        if Assigned(FOnHide) then
          FOnHide(Self);
      end;
    end;
    for i:= 0 to FWidgets.Count - 1 do
      TWidgetWindow(FWidgets[i]).SetVisible(Value);
  end;
end;

procedure TWidgetWindow.Show;
begin
  SetVisible(True);
end;

procedure TWidgetWindow.Hide;
begin
  SetVisible(False);
end;

function TWidgetWindow.SetRender(const FileName: String; Render: String): TWidgetBitmap;
var
  i: Integer;
  p: PRenderItem;
begin
  i := RenderIndexOf(Render);
  if (i >= 0) then
  begin
    PRenderItem(FRenders[i])^.Bitmap.Free;
    PRenderItem(FRenders[i])^.Bitmap := TWidgetBitmap.Create(FileName);
  end else
  begin
    New(p);
    p^.Name := Render;
    p^.Bitmap := TWidgetBitmap.Create(FileName);
    i := FRenders.Add(p);
  end;
  Result := PRenderItem(FRenders[i])^.Bitmap;
end;

function TWidgetWindow.SetRender(Stream: TStream; Render: String): TWidgetBitmap;
var
  i: Integer;
  p: PRenderItem;
begin
  i := RenderIndexOf(Render);
  if (i >= 0) then
  begin
    PRenderItem(FRenders[i])^.Bitmap.Free;
    PRenderItem(FRenders[i])^.Bitmap := TWidgetBitmap.Create(Stream);
  end else
  begin
    New(p);
    p^.Name := Render;
    p^.Bitmap := TWidgetBitmap.Create(Stream);
    i := FRenders.Add(p);
  end;
  Result := PRenderItem(FRenders[i])^.Bitmap;
end;

function TWidgetWindow.GetRender(Render: TWidgetBitmap): String;
var
  i: Integer;
begin
  Result := '';
  i := RenderIndexOf(Render);
  if (i >= 0) then
    Result := PRenderItem(FRenders[i])^.Name;
end;

function TWidgetWindow.GetRender(Render: String): TWidgetBitmap;
var
  i: Integer;
begin
  Result := NIL;
  i := RenderIndexOf(Render);
  if (i >= 0) then
    Result := PRenderItem(FRenders[i])^.Bitmap;
end;

procedure TWidgetWindow.RemoveRender(Render: String);
var
  i: Integer;
begin
  i := RenderIndexOf(Render);
  if (i >= 0) then
  begin
    PRenderItem(FRenders[i])^.Bitmap.Free;
    Dispose(PRenderItem(FRenders[i]));
    FRenders.Delete(i);
  end;
end;

procedure TWidgetWindow.RemoveRender(Render: TWidgetBitmap);
var
  i: Integer;
begin
  i := RenderIndexOf(Render);
  if (i >= 0) then
  begin
    PRenderItem(FRenders[i])^.Bitmap.Free;
    Dispose(PRenderItem(FRenders[i]));
    FRenders.Delete(i);
  end;
end;

procedure TWidgetWindow.RemoveAllRenders;
var
  i: Integer;
begin
  for i:= 0 to FRenders.Count - 1 do
  begin
    PRenderItem(FRenders[i])^.Bitmap.Free;
    Dispose(PRenderItem(FRenders[i]));  
  end;
  FRenders.Clear;
end;

procedure TWidgetWindow.UseRenderAsSize(Render: TWidgetBitmap);
begin
  FRect.Right := FRect.Left + Render.Width;
  FRect.Bottom := FRect.Top + Render.Height;
end;

procedure TWidgetWindow.UseRenderAsSize(Render: String);
var
  Bitmap: TWidgetBitmap;
begin
  Bitmap := GetRender(Render);
  if (Bitmap <> NIL) then
    UseRenderAsSize(Bitmap);
end;

function TWidgetWindow.RenderIndexOf(const Render: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i:= 0 to FRenders.Count - 1 do
  if SameText(Render, PRenderItem(FRenders[i])^.Name) then
  begin
    Result := i;
    Break;
  end;
end;

function TWidgetWindow.RenderIndexOf(Render: TWidgetBitmap): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i:= 0 to FRenders.Count - 1 do
  if (Render = PRenderItem(FRenders[i])^.Bitmap) then
  begin
    Result := i;
    Break;
  end;
end;

function TWidgetWindow.GetClientRect: TRect;
begin
  Result.Left := 0;
  Result.Top := 0;
  Result.Right := FRect.Right - FRect.Left;
  Result.Bottom := FRect.Bottom - FRect.Top;
end;

function TWidgetWindow.TimerIndexOf(const Name: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i:= 0 to FTimers.Count - 1 do
  if SameText(PTimerItem(FTimers[i])^.Name, Name) then
  begin
    Result := i;
    Break;
  end;
end;

function TWidgetWindow.GetTimerStarted(const Name: String): Boolean;
var
  i: Integer;
begin
  i := TimerIndexOf(Name);
  if (i >= 0) then
    Result := PTimerItem(FTimers[i])^.Runing
  else
    Result := False;
end;

function TWidgetWindow.GetTimerInterval(const Name: String): DWord;
var
  i: Integer;
begin
  i := TimerIndexOf(Name);
  if (i >= 0) then
    Result := PTimerItem(FTimers[i])^.Interval
  else
    Result := 0;
end;

procedure TWidgetWindow.RemoveTimer(const Name: String);
var
  i: Integer;
begin
  i := TimerIndexOf(Name);
  if (i >= 0) then
  begin
    KillTimer(FHandle, PTimerItem(FTimers[i])^.Index);
    Dispose(PTimerItem(FTimers[i]));
    FTimers.Delete(i);
  end;
end;

procedure TWidgetWindow.RemoveAllTimers;
var
  i: Integer;
begin
  for i:= 0 to FTimers.Count - 1 do
  begin
    KillTimer(FHandle, PTimerItem(FTimers[i])^.Index);
    Dispose(PTimerItem(FTimers[i]));
  end;
  FTimers.Clear;
end;

procedure TWidgetWindow.SetTimer(const Name: String; Interval: DWord;
  OnTimer: TWidgetEventTimer; Start: Boolean);
var
  i: Integer;
  p: PTimerItem;
begin
  i := TimerIndexOf(Name);
  if (i < 0) then
  begin
    New(p);
    p^.Index := FTimers.Add(p);
    p^.Name := Name;
    p^.Runing := False;
  end else
    p := FTimers[i];
  if p^.Runing then
    KillTimer(FHandle, p^.Index);
  p^.Interval := Interval;
  p^.OnTimer := OnTimer;
  p^.Runing := Start;
  if Start then
    Windows.SetTimer(FHandle, p^.Index, p^.Interval, NIL);
end;

procedure TWidgetWindow.SetTimer(const Name: String; Interval: DWord);
var
  i: Integer;
begin
  i := TimerIndexOf(Name);
  if (i >= 0) then
    SetTimer(Name, Interval, PTimerItem(FTimers[i])^.OnTimer, False);
end;

procedure TWidgetWindow.SetTimer(const Name: String; OnTimer: TWidgetEventTimer);
var
  i: Integer;
begin
  i := TimerIndexOf(Name);
  if (i >= 0) then
    SetTimer(Name, PTimerItem(FTimers[i])^.Interval, OnTimer, False);
end;

procedure TWidgetWindow.SetTimerState(const Name: String; Start: Boolean);
var
  i: Integer;
begin
  i := TimerIndexOf(Name);
  if (i >= 0) and (PTimerItem(FTimers[i])^.Runing <> Start) then
  begin
    if PTimerItem(FTimers[i])^.Runing then
      KillTimer(FHandle, PTimerItem(FTimers[i])^.Index);
    PTimerItem(FTimers[i])^.Runing := Start;
    if PTimerItem(FTimers[i])^.Runing then
      Windows.SetTimer(FHandle, PTimerItem(FTimers[i])^.Index, PTimerItem(FTimers[i])^.Interval, NIL);
  end;
end;

initialization
  WidgetWindowClass.lpfnWndProc := @DefWindowProc;
  WidgetWindowClass.hCursor := LoadCursor(0, IDC_ARROW);
  WidgetWindowClass.hInstance := HInstance;
  Windows.RegisterClass(WidgetWindowClass);
  WidgetHideWindow := CreateWindow(WidgetWindowClass.lpszClassName, '',
    WS_POPUP, 0, 0, 0, 0, 0, 0, HInstance, NIL);
  WidgetApplication := TWidgetApplication.Create;
  
finalization
  DestroyWindow(WidgetHideWindow);
  WidgetApplication.Free;

end.
