unit DWidgetGraphics;

interface

uses
  Windows, SysUtils, Classes, Graphics, Math, ActiveX, GDIPAPI;

type
  PARGB = ^TARGB;
  TARGB = packed record
  case Integer of
    0: (c: ARGB);
    1: (b, g, r, a: Byte);
  end;

  PARGBArray = ^TARGBArray;
  TARGBArray = array[0..(MAXINT div SizeOf(TARGB) - 1)] of TARGB;

  TWidgetBitmap = class
  private
    FWidth: Integer;
    FHeight: Integer;
    FBitmap: GpBitmap;
    FLockBitsCount: Integer;
    FBitmapData: TBitmapData;
    FBits: Pointer;

    function GetPixels(X, Y: Integer): Pointer;
    function GetSize: Integer;
    function GetRect: TRect;

    procedure AfterCreate;
  public
    constructor Create(Width, Height: Integer); overload;
    constructor Create(const FileName: String); overload;
    constructor Create(Stream: TStream); overload;
    destructor Destroy; override;

    procedure LockBits;
    procedure UnLockBits;

    property Width: Integer read FWidth;
    property Height: Integer read FHeight;
    property Rect: TRect read GetRect;
    property Pixels[X, Y: Integer]: Pointer read GetPixels; default;
    property Bits: Pointer read FBits;
    property Size: Integer read GetSize;
    property Bitmap: GpBitmap read FBitmap;
  end;

  TWidgetDIB = class
  private
    FWidth: Integer;
    FHeight: Integer;
    FBits: Pointer;
    FDC: HDC;
    FhBitmap: HBITMAP;
    FhOldBitmap: HBITMAP;
    FBitmapInfo: TBitmapInfo;

    FContext: GpGraphics;
    FBitmap: GpBitmap;
    FBrush: GpBrush;
    
    function GetPixels(X, Y: Integer): Pointer;
    function GetSize: Integer;
    function GetRect: TRect;

    function GetBrushColor: ARGB;
    procedure SetBrushColor(Color: ARGB);
  public
    constructor Create(Width, Height: Integer);
    destructor Destroy; override;
    
    procedure Assign(DIB: TWidgetDIB);

    procedure New(Width, Height: Integer);
    procedure Dispose;

    procedure Fill; overload;
    procedure Fill(ARect: TRect); overload;

    procedure Copy(DestRect: TRect; Src: TWidgetDIB; SrcRect: TRect);

    procedure Draw(DestRect: TRect; Src: TWidgetDIB; SrcRect: TRect);
    procedure DrawBitmap(DestRect: TRect; Src: GpBitmap; SrcRect: TRect);
    procedure DrawTo(DC: HDC; DestRect, SrcRect: TRect);

    procedure Premultiply(ARect: TRect);

    property Width: Integer read FWidth;
    property Height: Integer read FHeight;
    property Pixels[X, Y: Integer]: Pointer read GetPixels; default;
    property Bits: Pointer read FBits;
    property Size: Integer read GetSize;
    property Rect: TRect read GetRect;
    property DC: HDC read FDC;
    property hBitmap: HBITMAP read FhBitmap;
    property BrushColor: ARGB read GetBrushColor write SetBrushColor;
    property Context: GpGraphics read FContext;
    property Bitmap: GpBitmap read FBitmap;
    property Brush: GpBrush read FBrush;
  end;

implementation

var
  StartupInput: TGDIPlusStartupInput;
  GDIPlusToken: DWORD;

{ TWidgetDIB }

constructor TWidgetDIB.Create(Width, Height: Integer);
begin
  inherited Create;
  FWidth := 0;
  FHeight := 0;
  FDC := 0;
  FBits := NIL;
  FContext := NIL;
  FBitmap := NIL;
  GdipCreateSolidFill(ARGBMake(0, 0, 0), FBrush);
  New(Width, Height);
end;

destructor TWidgetDIB.Destroy;
begin
  Dispose;
  GdipDeleteBrush(FBrush);
  inherited;
end;

procedure TWidgetDIB.Assign(DIB: TWidgetDIB);
begin
  New(DIB.FWidth, DIB.FHeight);
  Move(DIB.FBits^, FBits^, GetSize);
end;

procedure TWidgetDIB.Dispose;
begin
  if (GetSize > 0) then
  begin
    GdipDeleteGraphics(FContext);
    GdipDisposeImage(FBitmap);
    SelectObject(FDC, FhOldBitmap);
    DeleteDC(FDC);
    DeleteObject(FhBitmap);
    FWidth := 0;
    FHeight := 0;
    FDC := 0;
    FBits := NIL;
  end;
end;

procedure TWidgetDIB.New(Width, Height: Integer);
begin
  if (FWidth = Width) and (FHeight = Height) then Exit;

  Dispose;
  
  FWidth := Width;
  FHeight := Height;

  with FBitmapInfo.bmiHeader do
  begin
    biSize := SizeOf(TBitmapInfoHeader);
    biWidth := FWidth;
    biHeight := FHeight;
    biPlanes := 1;
    biBitCount := 32;
    biCompression := BI_RGB;
    biSizeImage := GetSize;
    biXPelsPerMeter := 0;
    biYPelsPerMeter := 0;
    biClrUsed := 0;
    biClrImportant := 0;
  end;

  FDC := CreateCompatibleDC(0);
  FhBitmap := CreateDIBSection(FDC, FBitmapInfo, DIB_RGB_COLORS, FBits, 0, 0);
  FhOldBitmap := SelectObject(FDC, FhBitmap);

  GdipCreateBitmapFromScan0(FWidth, FHeight, FWidth * 4, PixelFormat32bppARGB, FBits, FBitmap);
  GdipCreateFromHDC(FDC, FContext);
  GdipSetSmoothingMode(FContext, SmoothingModeAntiAlias);
end;

function TWidgetDIB.GetRect: TRect;
begin
  Result.Left := 0;
  Result.Top := 0;
  Result.Right := FWidth;
  Result.Bottom := FHeight;
end;

function TWidgetDIB.GetBrushColor: ARGB;
begin
  GdipGetSolidFillColor(FBrush, Result);
end;

procedure TWidgetDIB.SetBrushColor(Color: ARGB);
begin
  GdipSetSolidFillColor(FBrush, Color);
end;

function TWidgetDIB.GetPixels(X, Y: Integer): Pointer;
begin
  Result := Pointer(Integer(FBits) + (FWidth * (FHeight - 1 - Y) + X) * 4);
end;

function TWidgetDIB.GetSize: Integer;
begin
  Result := FWidth * FHeight * 4;
end;

procedure TWidgetDIB.Fill(ARect: TRect);
begin
  GdipSetCompositingMode(FContext, CompositingModeSourceCopy);
  GdipFillRectangle(FContext, FBrush, ARect.Left, ARect.Top, ARect.Right - ARect.Left,
    ARect.Bottom - ARect.Top);
end;

procedure TWidgetDIB.Fill;
begin
  Fill(GetRect);
end;

procedure TWidgetDIB.Copy(DestRect: TRect; Src: TWidgetDIB; SrcRect: TRect);
begin
  GdipSetCompositingMode(FContext, CompositingModeSourceCopy);
  GdipDrawImageRectRect(FContext, Src.Bitmap, DestRect.Left,
    DestRect.Top, DestRect.Right - DestRect.Left, DestRect.Bottom - DestRect.Top,
    SrcRect.Left, SrcRect.Top, SrcRect.Right - SrcRect.Left,
    SrcRect.Bottom - SrcRect.Top, UnitPixel, NIL, NIL, NIL);
end;

procedure TWidgetDIB.DrawTo(DC: HDC; DestRect, SrcRect: TRect);
var
  Context: GpGraphics;
begin
  GdipCreateFromHDC(DC, Context);
  GdipSetSmoothingMode(Context, SmoothingModeAntiAlias);
  GdipSetCompositingMode(Context, CompositingModeSourceOver);
  GdipDrawImageRectRect(Context, FBitmap, DestRect.Left,
    DestRect.Top, DestRect.Right - DestRect.Left, DestRect.Bottom - DestRect.Top,
    SrcRect.Left, SrcRect.Top, SrcRect.Right - SrcRect.Left,
    SrcRect.Bottom - SrcRect.Top, UnitPixel, NIL, NIL, NIL);
  GdipDeleteGraphics(Context);
end;

procedure TWidgetDIB.Draw(DestRect: TRect; Src: TWidgetDIB; SrcRect: TRect);
begin
  GdipSetCompositingMode(FContext, CompositingModeSourceOver);
  GdipDrawImageRectRect(FContext, Src.Bitmap, DestRect.Left,
    DestRect.Top, DestRect.Right - DestRect.Left, DestRect.Bottom - DestRect.Top,
    SrcRect.Left, SrcRect.Top, SrcRect.Right - SrcRect.Left,
    SrcRect.Bottom - SrcRect.Top, UnitPixel, NIL, NIL, NIL);
end;

procedure TWidgetDIB.DrawBitmap(DestRect: TRect; Src: GpBitmap; SrcRect: TRect);
begin
  GdipSetCompositingMode(FContext, CompositingModeSourceOver);
  GdipDrawImageRectRect(FContext, Src, DestRect.Left, DestRect.Top,
    DestRect.Right - DestRect.Left, DestRect.Bottom - DestRect.Top,
    SrcRect.Left, SrcRect.Top, SrcRect.Right - SrcRect.Left,
    SrcRect.Bottom - SrcRect.Top, UnitPixel, NIL, NIL, NIL);
end;

procedure TWidgetDIB.Premultiply(ARect: TRect);
var
  x, y: Integer;
  p: PARGB;
begin
  for y:= ARect.Top to ARect.Bottom - 1 do
  for x:= ARect.Left to ARect.Right - 1 do
  begin
    p := GetPixels(x, y);
    if (p^.a = 0) then
      p^.a := 255;
  end;
end;

{ TWidgetBitmap }

constructor TWidgetBitmap.Create(Width, Height: Integer);
begin
  inherited Create;
  GdipCreateBitmapFromScan0(Width, Height, Width * 4, PixelFormat32bppARGB, NIL, FBitmap);
  AfterCreate;
end;

constructor TWidgetBitmap.Create(Stream: TStream);
var
  Str: TMemoryStream;
  iStr: IStream;
begin
  inherited Create;
  Str := TMemoryStream.Create;
  Str.CopyFrom(Stream, Stream.Size - Stream.Position);
  Str.Position := 0;
  iStr := TStreamAdapter.Create(Str, soOwned) as IStream;
  GdipCreateBitmapFromStream(iStr, FBitmap);
  AfterCreate;
end;

constructor TWidgetBitmap.Create(const FileName: String);
begin
  inherited Create;
  GdipCreateBitmapFromFile(PWCHAR(WideString(FileName)), FBitmap);
  AfterCreate;
end;

destructor TWidgetBitmap.Destroy;
begin
  if (FLockBitsCount > 0) then
  begin
    FLockBitsCount := 1;
    UnLockBits;
  end;
  GdipDisposeImage(FBitmap);
  inherited;
end;

procedure TWidgetBitmap.AfterCreate;
var
  v: DWORD;
begin
  GdipGetImageWidth(FBitmap, v);
  FWidth := Integer(v);
  GdipGetImageHeight(FBitmap, v);
  FHeight := Integer(v);
end;

function TWidgetBitmap.GetPixels(X, Y: Integer): Pointer;
begin
  Result := Pointer(Integer(FBits) + (FWidth * (FHeight - 1 - Y) + X) * 4);
end;

function TWidgetBitmap.GetRect: TRect;
begin
  Result.Left := 0;
  Result.Top := 0;
  Result.Right := FWidth;
  Result.Bottom := FHeight;
end;

function TWidgetBitmap.GetSize: Integer;
begin
  Result := FWidth * FHeight * 4;
end;

procedure TWidgetBitmap.LockBits;
var
  R: TGPRect;
begin
  if (FLockBitsCount = 0) then
  begin
    R.X := 0;
    R.Y := 0;
    R.Width := FWidth;
    R.Height := FHeight;
    GdipBitmapLockBits(FBitmap, @R, ImageLockModeRead or ImageLockModeWrite,
      PixelFormat32bppARGB, @FBitmapData);
    FBits := FBitmapData.Scan0;
  end;
  inc(FLockBitsCount);
end;

procedure TWidgetBitmap.UnLockBits;
begin
  if (FLockBitsCount = 1) then
    GdipBitmapUnlockBits(FBitmap, @FBitmapData);
  dec(FLockBitsCount);
end;

initialization
  StartupInput.DebugEventCallback := NIL;
  StartupInput.SuppressBackgroundThread := False;
  StartupInput.SuppressExternalCodecs := False;
  StartupInput.GdiplusVersion := 1;
  GdiplusStartup(GDIPlusToken, @StartupInput, NIL);

finalization
  GdiplusShutdown(GDIPlusToken);

end.
